from asyncio import wait, get_event_loop, Semaphore
from aiohttp import ClientSession, TCPConnector, InvalidURL


class Crawler:
    def __init__(self, config):
        self._entrance_urls = config.get('entrance_urls', None)
        self._init_clean = config.get('init_clean', None)
        if not self._entrance_urls or not self._init_clean:
            raise ValueError('The entrance_urls or the init_clean is none!')
        self._depth_clean = config.get('depth_clean', None)
        self._header = config.get('header', None)
        self._loop = get_event_loop()
        self._session = None
        self._semaphore = Semaphore(500, loop=self._loop)

    def run(self):
        self._loop.run_until_complete(self._start())
        self._loop.close()

    async def _start(self):
        conn = TCPConnector(limit=None, loop=self._loop)
        async with ClientSession(connector=conn) as session:
            self._session = session
            task_list = map(self._init_spider, self._entrance_urls)
            await wait(task_list, loop=self._loop)

    async def _init_spider(self, url):
        async def func(resp):
            depth_urls, download_list = await self._init_clean(resp=resp)
            task_list = []
            if depth_urls:
                task_list += map(self._depth_spider, depth_urls)
            if download_list:
                task_list += download_list
            if task_list:
                await wait(task_list, loop=self._loop)
        await self._request(url=url, func=func)

    async def _depth_spider(self, url):
        async def func(resp):
            download_list = await self._depth_clean(resp=resp)
            if download_list:
                await wait(download_list, loop=self._loop)
        await self._request(url=url, func=func)

    async def download(self, url, filename):
        async def func(resp):
            with open(filename, 'wb') as fd:
                while True:
                    chunk = await resp.content.read(1024)
                    if not chunk:
                        break
                    fd.write(chunk)
        await self._request(url=url, func=func)

    async def _request(self, url, func):
        try:
            with (await self._semaphore):
                async with self._session.get(url, headers=self._header) as resp:
                    await func(resp)
        except TimeoutError:
            print(f'Timeout -- {url}')
        except InvalidURL:
            print(f'Error URL -- {url}')
